"""The listener module contains the main Listener class which interacts with the TCI server."""

import asyncio
from asyncio.exceptions import CancelledError
import websockets

from . import tci

class Listener:
    """The Listener class interacts with the TCI server by listening for parameter updates.
    A sender task also passes formatted command strings & data packets to the server.
    Parameter and data stream callbacks can be registered to be notified of changes of interest,
    as many commands may be ignorable in certain use cases.
    """

    def __init__(self, uri):
        self.uri = uri
        self._tci_param_listeners = {}
        self._tci_data_listeners = {}
        self._tci_send = None
        self._launch_task = None
        self._connected_event = None
        self._ready_event = None

    @staticmethod
    def _convert_type(val):
        """Converts a string to int, float, or bool if possible, but returning the string if not."""
        try:
            return int(val)
        except (ValueError, TypeError):
            pass

        try:
            return float(val)
        except (ValueError, TypeError):
            pass

        if val.upper() == "TRUE":
            return True
        if val.upper() == "FALSE":
            return False

        return val

    def add_param_listener(self, param, callback):
        """Registers a callback to be notified of a particular parameter change.

        The callback signature is (param_name, rx, sub_rx, params).
        The special param "*" can be used to register a listener for all parameters.
        """
        if param not in self._tci_param_listeners:
            self._tci_param_listeners[param] = []
        l = self._tci_param_listeners[param]
        if callback not in l:
            l += [callback]

    def remove_param_listener(self, param, callback):
        """Removes a parameter callback from the notification list"""
        l = self._tci_param_listeners[param]
        if callback in l:
            l.remove(callback)

    def add_data_listener(self, data_type, callback):
        """Registers a callback to be notified when a particular type of data packet is received.

        The callback signature is (packet).
        The special data_type "*" can be used to register a listener for all data types.
        """
        if data_type not in self._tci_data_listeners:
            self._tci_data_listeners[data_type] = []
        l = self._tci_data_listeners[data_type]
        if callback not in l:
            l += [callback]

    def remove_data_listener(self, data_type, callback):
        """Removes a data callback from the notification list."""
        l = self._tci_data_listeners[data_type]
        if callback in l:
            l.remove(callback)

    def _get_param_listeners(self, item):
        """Retrieves the list of all parameter callbacks to notify for a particular parameter."""
        res = []
        if item in self._tci_param_listeners:
            res += self._tci_param_listeners[item]
        if "*" in self._tci_param_listeners:
            res += self._tci_param_listeners["*"]
        return res

    def _get_data_listeners(self, item):
        """Retrieves the list of all data callbacks to notify for a particular data type."""
        res = []
        if item in self._tci_data_listeners:
            res += self._tci_data_listeners[item]
        if "*" in self._tci_data_listeners:
            res += self._tci_data_listeners["*"]
        return res

    def _schedule_callback(self, callback, *callback_args):
        """Schedules a notification callback, ensuring that the status is checked when complete."""
        task = asyncio.create_task(callback(*callback_args))
        task.add_done_callback(lambda task: task.result())

    async def _listen_main(self, ws):
        """Coroutine that receives from the server and schedules data/parameter callbacks."""
        while True:
            status = await ws.recv()

            if isinstance(status, bytes):
                packet = tci.TciDataPacket.from_buf(status)
                for callback in self._get_data_listeners(packet.data_type):
                    self._schedule_callback(callback, packet)
                continue

            parts = status.strip(";").split(":", 1)
            cmd_name = parts[0].upper()
            if cmd_name not in tci.COMMANDS:
                raise ValueError(f'Command {cmd_name} unrecognized')

            cmd_info = tci.COMMANDS[cmd_name]
            expected_params = cmd_info.total_params()

            if expected_params == 0:
                if cmd_info.name == "READY":
                    self._ready_event.set()
                for callback in self._get_param_listeners(cmd_info.name):
                    self._schedule_callback(callback, cmd_info.name, None, None, None)
                continue

            if len(parts) != 2:
                raise ValueError(f'Command {cmd_name} should have parameters, but none received.')

            cmd_params = [Listener._convert_type(v) for v in parts[1].split(",")]
            param_cnt = len(cmd_params)
            if expected_params not in (-1, param_cnt):
                raise ValueError(f'Command {cmd_name} should have {expected_params} params, received {param_cnt}')

            param_rx = None
            if cmd_info.has_rx:
                param_rx = cmd_params.pop(0)
                param_cnt -= 1

            param_sub_rx = None
            if cmd_info.has_sub_rx:
                param_sub_rx = cmd_params.pop(0)
                param_cnt -= 1

            if param_cnt == 1:
                cmd_params = cmd_params[0]
            elif param_cnt == 0:
                cmd_params = None

            for callback in self._get_param_listeners(cmd_info.name):
                self._schedule_callback(callback, cmd_info.name, param_rx, param_sub_rx, cmd_params)

    async def _sender_main(self, ws):
        """Coroutine that sends commands and data packets to the server."""
        while True:
            msg = await self._tci_send.get()
            await ws.send(msg)

    async def _launch_tasks(self):
        """Coroutine that initiates connection and creates listener/sender tasks."""
        try:
            async with websockets.connect(self.uri) as ws:
                listen_task = asyncio.create_task(self._listen_main(ws))
                sender_task = asyncio.create_task(self._sender_main(ws))
                self._connected_event.set()
                try:
                    await listen_task
                    await sender_task
                except CancelledError:
                    listen_task.cancel()
                    sender_task.cancel()
        except CancelledError:
            pass

    async def send(self, data):
        """Coroutine to enqueue data for sending, ensuring it reaches the queue."""
        await self._tci_send.put(data)

    def send_nowait(self, data):
        """Enqueue data for sending without ensuring it reaches the queue."""
        self._tci_send.put_nowait(data)

    async def start(self, timeout=3.0):
        """Coroutine called to start the connection and listener/sender tasks."""
        if self._launch_task is not None and not self._launch_task.done():
            return
        self._tci_send = asyncio.Queue()
        self._connected_event = asyncio.Event()
        self._ready_event = asyncio.Event()
        self._launch_task = asyncio.create_task(self._launch_tasks())
        async def _await_conn():
            await self._connected_event.wait()
        done, _ = await asyncio.wait([asyncio.create_task(_await_conn())], timeout=timeout)
        if len(done) == 0:
            self._launch_task.cancel()
            try:
                await self._launch_task
            except CancelledError:
                pass
            except Exception as exc:
                raise TimeoutError(f'Connected event not received after {timeout} sec.') from exc
            raise TimeoutError(f'Connected event not received after {timeout} sec.')

    def shutdown(self):
        """Cancels communication tasks and shut down connection."""
        self._launch_task.cancel()

    async def ready(self, timeout=3.0):
        """Coroutine to verify initial synchronization is complete before continuing."""
        async def _await_ready():
            await self._ready_event.wait()
        done, _ = await asyncio.wait([asyncio.create_task(_await_ready())], timeout=timeout)
        if len(done) == 0:
            self._launch_task.cancel()
            try:
                await self._launch_task
            except CancelledError:
                pass
            except Exception as exc:
                raise TimeoutError(f'Ready event not received after {timeout} sec.') from exc
            raise TimeoutError(f'Ready event not received after {timeout} sec.')

    async def wait(self):
        """Coroutine used to wait for connection completion."""
        await self._launch_task



"""The tci module contains the constants, classes, and enums used to format commands and interact
with TCI audio and data streams.
"""

from enum import IntEnum
import struct

class TciCommand:
    """TciCommand instances define the individual commands that can be sent to or received from a
    TCI server and are used to check parameters and produce command strings.  The COMMANDS dict
    contains preconfigured TciCommand instances for currently-known commands retrievable by their
    name as key.
    """

    def __init__(self, name, readable = True, writeable = True,
                 has_rx = False, has_sub_rx = False, param_count = 1):
        self.name = name
        self.readable = readable
        self.writeable = writeable
        self.has_rx = has_rx
        self.has_sub_rx = has_sub_rx
        self.param_count = param_count

    def total_params(self):
        """Return the total number of parameters that should be present in a command string."""
        params = self.param_count
        if self.has_rx:
            params += 1
        if self.has_sub_rx:
            params += 1
        return params

    def prepare_string(self, action, rx = None, sub_rx = None, params = (), check_params = True):
        """Prepares a command string for sending using the provided values & parameters."""
        uc_command = self.name.upper()

        if action == TciCommandSendAction.READ and not self.readable:
            raise ValueError(f"Command {uc_command} not readable.")

        if action == TciCommandSendAction.WRITE and not self.writeable:
            raise ValueError(f"Command {uc_command} not writeable.")

        if self.has_rx:
            try:
                rx = int(rx)
                assert rx >= 0
            except (TypeError, ValueError, AssertionError) as exc:
                raise ValueError(f"Command {uc_command} requires specifying applicable receiver number (nonnegative integer)") from exc

        if self.has_sub_rx:
            try:
                sub_rx = int(sub_rx)
                assert sub_rx  >= 0
            except (TypeError, ValueError, AssertionError) as exc:
                raise ValueError(f"Command {uc_command} requires specifying applicable sub-receiver/channel number (nonnegative integer)") from exc

        if check_params:
            if action == TciCommandSendAction.READ and self.param_count > 0:
                expected = self.param_count - 1
            else:
                expected = self.param_count

            if len(params) != expected:
                raise ValueError(f"Command {uc_command} requires {expected} additional parameters to {action.name}, {len(params)} given.")

        cmd_params = []
        if self.has_rx:
            cmd_params += [str(rx)]
        if self.has_sub_rx:
            cmd_params += [str(sub_rx)]
        cmd_params += [str(p) for p in params]

        if len(cmd_params) == 0:
            return f"{uc_command};"

        param_str = ",".join(cmd_params)
        return f"{uc_command}:{param_str};"

COMMANDS = {cmd.name: cmd for cmd in [
    # Initialization Type Commands - TCI Protocol 2.0 - Section 4.1
    # All these should be readable = False, writeable = False
    TciCommand("VFO_LIMITS",              readable = False, writeable = False, param_count = 2),
    TciCommand("IF_LIMITS",               readable = False, writeable = False, param_count = 2),
    TciCommand("TRX_COUNT",               readable = False, writeable = False),
    TciCommand("CHANNELS_COUNT",          readable = False, writeable = False),
    TciCommand("DEVICE",                  readable = False, writeable = False),
    TciCommand("RECEIVE_ONLY",            readable = False, writeable = False),
    TciCommand("MODULATIONS_LIST",        readable = False, writeable = False, param_count = -1),
    TciCommand("PROTOCOL",                readable = False, writeable = False, param_count = 2),
    TciCommand("READY",                   readable = False, writeable = False, param_count = 0),
    # Bidirectional Control Commands - TCI Protocol 2.0 - Section 4.2
    # All these should be readable = True, writeable = True
    TciCommand("START",                   readable = False, param_count = 0),
    TciCommand("STOP",                    readable = False, param_count = 0),
    TciCommand("DDS",                     has_rx = True),
    TciCommand("IF",                      has_rx = True, has_sub_rx = True),
    TciCommand("VFO",                     has_rx = True, has_sub_rx = True),
    TciCommand("MODULATION",              has_rx = True),
    TciCommand("TRX",                     has_rx = True), # has optional parameter for TCI audio only for sending
    TciCommand("TUNE",                    has_rx = True),
    TciCommand("DRIVE",                   has_rx = True),
    TciCommand("TUNE_DRIVE",              has_rx = True),
    TciCommand("RIT_ENABLE",              has_rx = True),
    TciCommand("XIT_ENABLE",              has_rx = True),
    TciCommand("SPLIT_ENABLE",            has_rx = True),
    TciCommand("RIT_OFFSET",              has_rx = True),
    TciCommand("XIT_OFFSET",              has_rx = True),
    TciCommand("RX_CHANNEL_ENABLE",       has_rx = True, has_sub_rx = True),
    TciCommand("RX_FILTER_BAND",          has_rx = True, param_count = 2),
    TciCommand("CW_MACROS_SPEED"),
    TciCommand("CW_MACROS_DELAY"),
    TciCommand("CW_KEYER_SPEED"),
    TciCommand("VOLUME"),
    TciCommand("MUTE"),
    TciCommand("RX_MUTE",                 has_rx = True),
    TciCommand("RX_VOLUME",               has_rx = True, has_sub_rx = True),
    TciCommand("RX_BALANCE",              has_rx = True, has_sub_rx = True),
    TciCommand("MON_VOLUME"),
    TciCommand("MON_ENABLE"),
    TciCommand("AGC_MODE",                has_rx = True),
    TciCommand("AGC_GAIN",                has_rx = True),
    TciCommand("RX_NB_ENABLE",            has_rx = True),
    TciCommand("RX_NB_PARAM",             has_rx = True, param_count = 2),
    TciCommand("RX_BIN_ENABLE",           has_rx = True),
    TciCommand("RX_NR_ENABLE",            has_rx = True),
    TciCommand("RX_ANC_ENABLE",           has_rx = True),
    TciCommand("RX_ANF_ENABLE",           has_rx = True),
    TciCommand("RX_APF_ENABLE",           has_rx = True),
    TciCommand("RX_DSE_ENABLE",           has_rx = True),
    TciCommand("RX_NF_ENABLE",            has_rx = True),
    TciCommand("LOCK",                    has_rx = True),
    TciCommand("SQL_ENABLE",              has_rx = True),
    TciCommand("SQL_LEVEL",               has_rx = True),
    TciCommand("DIGL_OFFSET"),
    TciCommand("DIGU_OFFSET"),
    # Unidirectional Control Commands - TCI Protocol 2.0 - Section 4.3
    # All these should be readable = False, writeable = True/False depending on semantics
    TciCommand("TX_ENABLE",               readable = False, writeable = False, has_rx = True),
    TciCommand("CW_MACROS_SPEED_UP",      readable = False),
    TciCommand("CW_MACROS_SPEED_DOWN",    readable = False),
    TciCommand("SPOT",                    readable = False, param_count = 5),
    TciCommand("SPOT_DELETE",             readable = False),
    TciCommand("IQ_SAMPLERATE",           readable = False),
    TciCommand("AUDIO_SAMPLERATE",        readable = False),
    TciCommand("IQ_START",                readable = False, has_rx = True, param_count = 0),
    TciCommand("IQ_STOP",                 readable = False, has_rx = True, param_count = 0),
    TciCommand("AUDIO_START",             readable = False, has_rx = True, param_count = 0),
    TciCommand("AUDIO_STOP",              readable = False, has_rx = True, param_count = 0),
    TciCommand("LINE_OUT_START",          readable = False, has_rx = True, param_count = 0),
    TciCommand("LINE_OUT_STOP",           readable = False, has_rx = True, param_count = 0),
    TciCommand("LINE_OUT_RECORDER_START", readable = False, has_rx = True),
    TciCommand("LINE_OUT_RECORDER_SAVE",  readable = False, has_rx = True),
    TciCommand("LINE_OUT_RECORDER_BREAK", readable = False, has_rx = True, param_count = 0),
    TciCommand("SPOT_CLEAR",              readable = False, param_count = 0),
    TciCommand("AUDIO_STREAM_SAMPLE_TYPE",readable = False),
    TciCommand("AUDIO_STREAM_CHANNELS",   readable = False),
    TciCommand("AUDIO_STREAM_SAMPLES",    readable = False),
    TciCommand("TX_STREAM_AUDIO_BUFFERING", readable = False),
    # Notification Commands - TCI Protocol 2.0 - Section 4.4
    # All these should be readable = False, writeable = False, but a few commands are mixed into this section
    TciCommand("CLICKED_ON_SPOT",         readable = False, writeable = False, param_count = 2),
    TciCommand("RX_CLICKED_ON_SPOT",      readable = False, writeable = False, has_rx = True, has_sub_rx = True, param_count = 2),
    TciCommand("TX_FOOTSWITCH",           readable = False, writeable = False, has_rx = True),
    TciCommand("TX_FREQUENCY",            readable = False, writeable = False),
    TciCommand("APP_FOCUS",               readable = False, writeable = False),
    TciCommand("SET_IN_FOCUS",            readable = False, param_count = 0),
    TciCommand("KEYER",                   readable = False, writeable = False, has_rx = True),
    TciCommand("RX_SENSORS_ENABLE",       readable = False, param_count = 2),
    TciCommand("TX_SENSORS_ENABLE",       readable = False, param_count = 2),
    TciCommand("RX_SENSORS",              readable = False, writeable = False, has_rx = True), # Deprecated in 2.0
    TciCommand("TX_SENSORS",              readable = False, writeable = False, has_rx = True, param_count = 4),
    # New Commands - TCI Protocol 2.0 - Section 4.5
    TciCommand("VFO_LOCK",                readable = False, writeable = False, has_rx = True, has_sub_rx = True),
    TciCommand("RX_CHANNEL_SENSORS",      readable = False, writeable = False, has_rx = True, has_sub_rx = True),
    # CW Macros - TCI Protocl 2.0 - Section 3.2.1
    TciCommand("CW_MACROS",               readable = False, has_rx = True),
    TciCommand("CW_TERMINAL",             readable = False),
    TciCommand("CW_MACROS_EMPTY",         readable = False, writeable = False, param_count = 0),
    TciCommand("CW_MSG",                  readable = False, has_rx = True, param_count = 3),
    TciCommand("CALLSIGN_SEND",           readable = False, writeable = False),
    TciCommand("CW_MACROS_STOP",          readable = False, param_count = 0),
    # Commands not documented in 1.9 but definitely encountered - TCI Protocol 1.6
    TciCommand("RX_ENABLE",               has_rx = True),
    TciCommand("CTCSS_ENABLE",            has_rx = True),
    TciCommand("CTCSS_MODE",              has_rx = True),
    TciCommand("CTCSS_RX_TONE",           has_rx = True),
    TciCommand("CTCSS_TX_TONE",           has_rx = True),
    TciCommand("CTCSS_LEVEL",             has_rx = True),
    # Commands not documented in 1.9 but no similar functionality - TCI Protocol 1.6
    TciCommand("ECODER_SWITCH_RX",        has_rx = True),
    TciCommand("ECODER_SWITCH_CHANNEL",   has_rx = True),
    # Commands not documented in 1.9 but probably superseded - TCI Protocol 1.6
    TciCommand("RX_SMETER",               writeable = False, has_rx = True, has_sub_rx = True),
    TciCommand("TX_POWER",                writeable = False),
    TciCommand("TX_SWR",                  writeable = False),
]}

class TciCommandSendAction(IntEnum):
    """TciCommandSendAction defines whether a parameter update is being requested (READ)
    or sent to the device (WRITE)."""
    READ = 0
    WRITE = 1

class TciStreamType(IntEnum):
    """TciStreamType defines the type of data stream contained in a data packet."""
    IQ_STREAM = 0
    RX_AUDIO_STREAM = 1
    TX_AUDIO_STREAM = 2
    TX_CHRONO = 3

class TciSampleType(IntEnum):
    """TciSampleType defines the type of individual samples of data contained in a data packet."""
    INT16 = 0
    INT24 = 1
    INT32 = 2
    FLOAT32 = 3

class TciDataPacket:
    """TciDataPacket instances are used to contain received data packets or define outgoing data packets.
    """

    def __init__(self, rx, sample_rate, data_format, codec, crc, length, data_type, channels, data):
        self.rx = rx
        self.sample_rate = sample_rate
        self.data_format = data_format
        self.codec = codec
        self.crc = crc
        self.length = length
        self.data_type = data_type
        self.channels = channels
        self.data = data

    @classmethod
    def from_buf(cls, buf):
        """Produces a TciDataPacket instance by decoding a raw recevied data buffer."""
        vals = struct.unpack_from("<8I", buf)
        rx = vals[0]
        sample_rate = vals[1]
        data_format = TciSampleType(vals[2])
        codec = vals[3]
        crc = vals[4]
        length = vals[5]
        data_type = TciStreamType(vals[6])
        channels = vals[7]
        offset = 8*4+8*4
        if length:
            data = buf[offset:]
        else:
            data = None
        return cls(rx, sample_rate, data_format, codec, crc, length, data_type, channels, data)

    def to_bytes(self):
        """Returns the raw bytes required to represent a TciDataPacket instance."""
        if self.data_format == TciSampleType.INT16:
            bytes_per_sample = 2
        elif self.data_format == TciSampleType.INT24:
            bytes_per_sample = 3
        else:
            bytes_per_sample = 4
        return struct.pack(f"<8I32x{bytes_per_sample*self.length}s", self.rx, self.sample_rate, self.data_format, self.codec, self.crc, self.length, self.data_type, self.channels, self.data)
